@@ -3,13 +3,15 @@ import fnmatch
 import logging
 from collections import Counter, defaultdict
 from datetime import datetime, timezone
-from itertools import chain, repeat, starmap
+from itertools import chain, islice, repeat, starmap
 from pathlib import Path
 from uuid import UUID
 
 from arkindex_cli.commands.export.utils import MANUAL_SOURCE, uuid_or_manual
 from arkindex_export import (
     Classification,
+    Dataset,
+    DatasetElement,
     Element,
     ElementPath,
     EntityType,
@@ -21,7 +23,7 @@ from arkindex_export import (
     open_database,
 )
 from arkindex_export.queries import list_children, list_parents
-from peewee import JOIN, fn
+from peewee import JOIN, ModelSelect, fn
 
 logger = logging.getLogger(__name__)
 
@@ -231,6 +233,31 @@ def metadata_columns(elements, load_parents=False):
     return columns
 
 
+def dataset_columns(elements: ModelSelect | None) -> list[str]:
+    # Count the DatasetElements by element ID
+    dataset_elements = DatasetElement.select(
+        DatasetElement.element_id,
+        fn.COUNT("*").alias("count"),
+    ).group_by(DatasetElement.element_id)
+
+    # Apply the element filters if there were any
+    if elements is not None:
+        dataset_elements = dataset_elements.where(
+            DatasetElement.element_id.in_(elements.select(Element.id))
+        )
+
+    # Find the maximum count across all elements
+    # This may be `None` when zero DatasetElements have been found
+    max_count = (
+        DatasetElement.select(fn.MAX(dataset_elements.c.count))
+        .from_(dataset_elements)
+        .scalar()
+        or 0
+    )
+
+    return ["id", "name", "set_name"] * max_count
+
+
 def element_classes(element_id, classification_worker_version=None):
     element_classifications = Classification.select(
         Classification.class_name, Classification.confidence
@@ -294,6 +321,9 @@ def element_transcriptions(element_id, transcriptions_worker_version=None):
             Transcription.confidence,
             Transcription.orientation,
             WorkerRun.worker_version_id,
+            fn.ROW_NUMBER()
+            .over(partition_by=([WorkerRun.worker_version_id]))
+            .alias("number"),
         )
         .where(Transcription.element_id == element_id)
         .join(
@@ -310,6 +340,19 @@ def element_transcriptions(element_id, transcriptions_worker_version=None):
     return transcriptions.namedtuples()
 
 
+def element_datasets(element_id: str) -> list[dict]:
+    return list(
+        DatasetElement.select(
+            DatasetElement.dataset_id.alias("id"),
+            Dataset.name,
+            DatasetElement.set_name,
+        )
+        .join(Dataset)
+        .where(DatasetElement.element_id == element_id)
+        .dicts()
+    )
+
+
 def is_present(csv_header, key):
     """
     Check if there are columns in the CSV header prefixed with either metadata, classification
@@ -326,6 +369,7 @@ def element_dict(
     with_parent_metadata=False,
     with_entities=False,
     with_transcriptions=False,
+    with_datasets=False,
     classes_columns=None,
     metadata_columns=None,
     entity_type_columns=None,
@@ -383,20 +427,23 @@ def element_dict(
     if with_transcriptions:
         transcriptions = element_transcriptions(item.id, transcriptions_worker_version)
         for transcription in transcriptions:
-            serialized_element["transcriptions"][
-                f"text_{transcription.worker_version or 'manual'}"
-            ].append(transcription.text)
-            serialized_element["transcriptions"][
-                f"confidence_{transcription.worker_version or 'manual'}"
-            ].append(transcription.confidence)
-            serialized_element["transcriptions"][
-                f"orientation_{transcription.worker_version or 'manual'}"
-            ].append(transcription.orientation)
+            # Transcription data goes by 3 (text, confidence, orientation)
+            key = f"{transcription.worker_version or 'manual'}"
+            serialized_element["transcriptions"][key].append(
+                {
+                    "text": transcription.text,
+                    "confidence": transcription.confidence,
+                    "orientation": transcription.orientation,
+                }
+            )
+    # datasets
+    if with_datasets:
+        serialized_element["datasets"] = element_datasets(item.id)
 
     return serialized_element
 
 
-def fill_columns(element_dict, header, row, k, columns, prefix, columns_offset=1):
+def fill_columns(element_dict, header, row, k, columns, prefix):
     """
     Fill rows with the appropriate metadata, classifications and entities values. This adds the
     values at the right position (in the correct columns) and fills columns with None when there
@@ -406,19 +453,90 @@ def fill_columns(element_dict, header, row, k, columns, prefix, columns_offset=1
     # Only keep the information for which there is a column in the CSV header
     columns = [column for column in columns if f"{prefix}_{column}" in header]
     column_counts = Counter(columns)
+
     for column, column_count in column_counts.items():
         if column not in element_dict[k]:
             row.extend(repeat(None, column_count))
         else:
-            for i in range(0, column_count, columns_offset):
+            for i in range(column_count):
                 row.append(
                     element_dict[k][column][i]
                     if i <= len(element_dict[k][column]) - 1
                     else None
                 )
+
     return row
 
 
+def fill_columns_with_dict(element_dict, header, column_prefix, element_dict_key):
+    """
+    A simpler version of fill_columns to support multiple columns per related object (like transcriptions)
+    Requires a list of related objects per worker versions (or manual)
+    Columns must have the structure {prefix}_{related_object_key}_{worker_version}
+    """
+    out = []
+
+    # Support multiple values per worker version
+    indexes = defaultdict(int)
+
+    for column in header:
+        # Skip every column not used by this related object
+        if not column.startswith(column_prefix):
+            continue
+
+        # Each column must hold the related object key and worker version
+        # Do not use .split as worker version could hold extra underscores
+        offset = len(column_prefix) + 1
+        key, _, worker_version = column[offset:].partition("_")
+
+        # Retrieve the value deep down into the element dict
+        try:
+            index = indexes[worker_version + key]
+            value = element_dict[element_dict_key][worker_version][index][key]
+        except (IndexError, KeyError):
+            value = None
+
+        # Add one value per column only
+        out.append(value)
+
+        # Increment the worker version counter for that key
+        indexes[worker_version + key] += 1
+
+    return out
+
+
+def fill_columns_list(
+    items: list[dict],
+    csv_header: list[str],
+    row: list,
+    columns: list[str],
+    column_prefix: str,
+) -> None:
+    """
+    From a `items` list of dicts, adds the values to `row` which match the columns of `columns`
+    that are found in the `csv_header` as `{column_prefix}_{column}`.
+    Each dict is expected to have keys that match column names.
+    The column names must be the same between each item, no UUIDs, no incrementing integers.
+    """
+    # Only consider columns that have not been filtered out from the final header
+    columns = [
+        column for column in columns if f"{column_prefix}_{column}" in csv_header
+    ]
+    # Get unique column names while preserving order. A set() does not preserve the order, so we use dict keys instead
+    unique_columns = list(dict.fromkeys(columns))
+
+    # Add the values for all the matched columns in every item, then pad with None until we reach the total columns
+    row.extend(
+        islice(
+            chain(
+                (item[column] for item in items for column in unique_columns),
+                repeat(None),
+            ),
+            len(columns),
+        )
+    )
+
+
 def element_row(
     element_dict,
     csv_columns,
@@ -426,6 +544,7 @@ def element_row(
     metadata_columns=None,
     entity_type_columns=None,
     transcription_columns=None,
+    dataset_columns=None,
 ):
     # Ensure the base (non metadata, classification or entity) information are in the correct order
     row = [
@@ -457,14 +576,19 @@ def element_row(
         )
     # Add transcriptions
     if transcription_columns:
-        row = fill_columns(
+        row += fill_columns_with_dict(
             element_dict,
             csv_columns,
+            column_prefix="transcription",
+            element_dict_key="transcriptions",
+        )
+    if dataset_columns:
+        fill_columns_list(
+            element_dict["datasets"],
+            csv_columns,
             row,
-            "transcriptions",
-            transcription_columns,
-            "transcription",
-            columns_offset=3,
+            dataset_columns,
+            "dataset",
         )
 
     return row
@@ -482,6 +606,7 @@ def build_csv_header(
     cl_columns=None,
     et_columns=None,
     tr_columns=None,
+    ds_columns=None,
 ):
     csv_header = [
         "id",
@@ -495,19 +620,17 @@ def build_csv_header(
         "created",
     ]
 
-    # Order in the CSV header: metadata, classifications, entities
-    if md_columns:
-        csv_header = csv_header + [f"metadata_{col_name}" for col_name in md_columns]
-    if cl_columns:
-        csv_header = csv_header + [
-            f"classification_{col_name}" for col_name in cl_columns
-        ]
-    if et_columns:
-        csv_header = csv_header + [f"entity_{col_name}" for col_name in et_columns]
-    if tr_columns:
-        csv_header = csv_header + [
-            f"transcription_{col_name}" for col_name in tr_columns
-        ]
+    # Add the columns for the related objects in a specific order
+    extra_columns = [
+        ("metadata", md_columns),
+        ("classification", cl_columns),
+        ("entity", et_columns),
+        ("transcription", tr_columns),
+        ("dataset", ds_columns),
+    ]
+    for prefix, columns in extra_columns:
+        if columns:
+            csv_header.extend([f"{prefix}_{col_name}" for col_name in columns])
 
     if output_header:
         csv_header = filter_columns(csv_header, output_header)
@@ -528,6 +651,7 @@ def run(
     with_parent_metadata: bool | None = False,
     with_entities: bool | None = False,
     with_transcriptions: bool | None = False,
+    with_datasets: bool | None = False,
     classification_worker_version: str | None = None,
     entities_worker_version: str | None = None,
     transcriptions_worker_version: str | None = None,
@@ -549,62 +673,69 @@ def run(
 
     elements = get_elements(parent, type, recursive)
 
+    # To allow the `*_columns` methods to avoid unnecessarily filters on elements when all elements are meant to be selected,
+    # we will pass `None` as the element query when no filters have been enabled.
+    column_elements = elements if parent or type else None
+
     md_columns = None
     cl_columns = None
     et_columns = None
     tr_columns = None
+    ds_columns = None
     if with_metadata:
         # Fetch all the metadata keys to build one CSV column by key
-        if not parent and not type:
-            md_columns = metadata_columns(
-                None,
-                load_parents=with_parent_metadata,
-            )
-        else:
-            md_columns = metadata_columns(
-                elements,
-                load_parents=with_parent_metadata,
-            )
+        md_columns = metadata_columns(
+            column_elements,
+            load_parents=with_parent_metadata,
+        )
     if with_classes:
-        if not parent and not type:
-            cl_columns = classes_columns(None)
-        else:
-            cl_columns = classes_columns(elements)
+        cl_columns = classes_columns(column_elements)
     if with_entities:
-        if not parent and not type:
-            et_columns = entity_type_columns(None, entities_worker_version)
-        else:
-            et_columns = entity_type_columns(elements, entities_worker_version)
+        et_columns = entity_type_columns(column_elements, entities_worker_version)
     if with_transcriptions:
-        if not parent and not type:
-            tr_columns = transcriptions_columns(None, transcriptions_worker_version)
-        else:
-            tr_columns = transcriptions_columns(elements, transcriptions_worker_version)
+        tr_columns = transcriptions_columns(
+            column_elements, transcriptions_worker_version
+        )
+    if with_datasets:
+        ds_columns = dataset_columns(column_elements)
 
     csv_header = build_csv_header(
-        output_header, md_columns, cl_columns, et_columns, tr_columns
+        output_header, md_columns, cl_columns, et_columns, tr_columns, ds_columns
     )
 
     # Disable classes, metadata and classification options if there are no corresponding columns in the header
-    if not is_present(csv_header, "entity"):
+    if not is_present(csv_header, "entity_"):
         if with_entities and output_header:
             logger.warning(
                 "After filtering columns based on the --field argument, no entities will be exported."
             )
         with_entities = False
-    if not is_present(csv_header, "classification"):
+    if not is_present(csv_header, "classification_"):
         if with_classes and output_header:
             logger.warning(
                 "After filtering columns based on the --field argument, no classifications will be exported."
             )
         with_classes = False
-    if not is_present(csv_header, "metadata"):
+    if not is_present(csv_header, "metadata_"):
         if with_metadata and output_header:
             logger.warning(
                 "After filtering columns based on the --field argument, no metadata will be exported."
             )
         with_metadata = False
         with_parent_metadata = False
+    if not is_present(csv_header, "transcription_"):
+        if with_transcriptions and output_header:
+            logger.warning(
+                "After filtering columns based on the --field argument, no transcriptions will be exported."
+            )
+        with_transcriptions = False
+    if not is_present(csv_header, "dataset_"):
+        if with_datasets and output_header:
+            logger.warning(
+                "After filtering columns based on the --field argument, no datasets will be exported."
+            )
+        with_datasets = False
+        ds_columns = None
 
     elements = (
         elements.join(Image, JOIN.LEFT_OUTER, on=[Image.id == Element.image_id])
@@ -625,6 +756,7 @@ def run(
                 with_parent_metadata,
                 with_entities,
                 with_transcriptions,
+                with_datasets,
                 classes_columns=cl_columns,
                 metadata_columns=md_columns,
                 entity_type_columns=et_columns,
@@ -634,7 +766,13 @@ def run(
             )
 
             item_row = element_row(
-                item_dict, csv_header, cl_columns, md_columns, et_columns, tr_columns
+                item_dict,
+                csv_header,
+                cl_columns,
+                md_columns,
+                et_columns,
+                tr_columns,
+                ds_columns,
             )
             writer.writerow(item_row)
 
@@ -692,6 +830,11 @@ def add_csv_parser(subcommands):
         type=uuid_or_manual,
         help="Only retrieve the transcriptions created by a specific worker version.",
     )
+    csv_parser.add_argument(
+        "--with-datasets",
+        action="store_true",
+        help="Retrieve element datasets.",
+    )
     csv_parser.add_argument(
         "-o",
         "--output",
