@@ -1,3 +1,4 @@
+from arkindex_cli.commands.worker.import_worker import add_import_parser
 from arkindex_cli.commands.worker.publish import add_publish_parser
 
 
@@ -8,6 +9,7 @@ def add_worker_parser(subcommands) -> None:
         help="Manage workers and their versions.",
     )
     subparsers = export_parser.add_subparsers()
+    add_import_parser(subparsers)
     add_publish_parser(subparsers)
 
     def subcommand_required(*args, **kwargs):
