@@ -1128,3 +1128,13 @@ def test_with_transcriptions(
     with (samples_dir / "transcriptions_output.json").open("r") as samples:
         data = json.loads(samples.read())[expected_output]
     assert rows == data
+
+
+def test_with_datasets(export_db_path, samples_dir):
+    with tempfile.NamedTemporaryFile(suffix=".csv") as f:
+        run(
+            export_db_path,
+            output_path=Path(f.name),
+            with_datasets=True,
+        )
+        assert f.read() == (samples_dir / "with_datasets.csv").read_bytes()
